@@ -78,6 +78,7 @@ typedef struct {
     pthread_cond_t cond;
     obj_io *queue;
     store_engine *e;
+    unsigned int depth; // queue depth
 } store_io_thread;
 
 typedef struct {
@@ -86,7 +87,6 @@ typedef struct {
     store_engine *e;
 } store_maint_thread;
 
-/* TODO: Array of FDs for JBOD support */
 struct store_engine {
     pthread_mutex_t mutex; /* covers internal stacks and variables */
     store_page *pages; /* directly addressable page list */
@@ -126,10 +126,21 @@ static _store_wbuf *wbuf_new(size_t size) {
 }
 
 static store_io_thread *_get_io_thread(store_engine *e) {
-    int tid;
+    int tid = -1;
+    long long int low = LLONG_MAX;
     pthread_mutex_lock(&e->mutex);
-    tid = (e->last_io_thread + 1) % e->io_threadcount;
-    e->last_io_thread = tid;
+    // find smallest queue. ignoring lock since being wrong isn't fatal.
+    // TODO: if average queue depth can be quickly tracked, can break as soon
+    // as we see a thread that's less than average, and start from last_io_thread
+    for (int x = 0; x < e->io_threadcount; x++) {
+        if (e->io_threads[x].depth == 0) {
+            tid = x;
+            break;
+        } else if (e->io_threads[x].depth < low) {
+                tid = x;
+            low = e->io_threads[x].depth;
+        }
+    }
     pthread_mutex_unlock(&e->mutex);
 
     return &e->io_threads[tid];
@@ -154,6 +165,12 @@ void extstore_get_stats(void *ptr, struct extstore_stats *st) {
     st->pages_free = e->page_free;
     st->pages_used = e->page_count - e->page_free;
     pthread_mutex_unlock(&e->mutex);
+    st->io_queue = 0;
+    for (int x = 0; x < e->io_threadcount; x++) {
+        pthread_mutex_lock(&e->io_threads[x].mutex);
+        st->io_queue += e->io_threads[x].depth;
+        pthread_mutex_unlock(&e->io_threads[x].mutex);
+    }
     // calculate bytes_fragmented.
     // note that open and yet-filled pages count against fragmentation.
     st->bytes_fragmented = st->pages_used * e->page_size -
@@ -169,7 +186,7 @@ void extstore_get_page_data(void *ptr, struct extstore_stats *st) {
 }
 
 const char *extstore_err(enum extstore_res res) {
-    char *rv = "unknown error";
+    const char *rv = "unknown error";
     switch (res) {
         case EXTSTORE_INIT_BAD_WBUF_SIZE:
             rv = "page_size must be divisible by wbuf_size";
@@ -183,6 +200,9 @@ const char *extstore_err(enum extstore_res res) {
         case EXTSTORE_INIT_PAGE_WBUF_ALIGNMENT:
             rv = "page_size and wbuf_size must be divisible by 1024*1024*2";
             break;
+        case EXTSTORE_INIT_TOO_MANY_PAGES:
+            rv = "page_count must total to < 65536. Increase page_size or lower path sizes";
+            break;
         case EXTSTORE_INIT_OOM:
             rv = "failed calloc for engine";
             break;
@@ -230,6 +250,7 @@ void *extstore_init(struct extstore_conf_file *fh, struct extstore_conf *cf,
     }
 
     e->page_size = cf->page_size;
+    uint64_t temp_page_count = 0;
     for (f = fh; f != NULL; f = f->next) {
         f->fd = open(f->file, O_RDWR | O_CREAT | O_TRUNC, 0644);
         if (f->fd < 0) {
@@ -240,10 +261,17 @@ void *extstore_init(struct extstore_conf_file *fh, struct extstore_conf *cf,
             free(e);
             return NULL;
         }
-        e->page_count += f->page_count;
+        temp_page_count += f->page_count;
         f->offset = 0;
     }
 
+    if (temp_page_count >= UINT16_MAX) {
+        *res = EXTSTORE_INIT_TOO_MANY_PAGES;
+        free(e);
+        return NULL;
+    }
+    e->page_count = temp_page_count;
+
     e->pages = calloc(e->page_count, sizeof(store_page));
     if (e->pages == NULL) {
         *res = EXTSTORE_INIT_OOM;
@@ -579,6 +607,12 @@ int extstore_submit(void *ptr, obj_io *io) {
         }
         tmp->next = io;
     }
+    // TODO: extstore_submit(ptr, io, count)
+    obj_io *tio = io;
+    while (tio != NULL) {
+        t->depth++;
+        tio = tio->next;
+    }
     pthread_mutex_unlock(&t->mutex);
 
     //pthread_mutex_lock(&t->mutex);
@@ -703,6 +737,7 @@ static void *extstore_io_thread(void *arg) {
                     break;
                 }
             }
+            me->depth -= i;
             me->queue = end->next;
             end->next = NULL;
         }
@@ -738,11 +773,22 @@ static void *extstore_io_thread(void *arg) {
                     }
                     pthread_mutex_unlock(&p->mutex);
                     if (do_op) {
+#ifdef __APPLE__
+                        ret = lseek(p->fd, SEEK_SET, p->offset + cur_io->offset);
+                        if (ret >= 0) {
+                            if (cur_io->iov == NULL) {
+                                ret = read(p->fd, cur_io->buf, cur_io->len);
+                            } else {
+                                ret = readv(p->fd, cur_io->iov, cur_io->iovcnt);
+                            }
+                        }
+#else
                         if (cur_io->iov == NULL) {
                             ret = pread(p->fd, cur_io->buf, cur_io->len, p->offset + cur_io->offset);
                         } else {
                             ret = preadv(p->fd, cur_io->iov, cur_io->iovcnt, p->offset + cur_io->offset);
                         }
+#endif
                     }
                     break;
                 case OBJ_IO_WRITE:
