@@ -6,6 +6,7 @@
 #include <errno.h>
 #include <poll.h>
 #include <ctype.h>
+#include <stdarg.h>
 
 #if defined(__sun)
 #include <atomic.h>
@@ -87,8 +88,8 @@ static int logger_thread_poll_watchers(int force_poll, int watcher);
 
 /* Logger GID's can be used by watchers to put logs back into strict order
  */
-static uint64_t logger_get_gid(void) {
-    static uint64_t logger_gid = 0;
+static uint64_t logger_gid = 0;
+uint64_t logger_get_gid(void) {
 #ifdef HAVE_GCC_64ATOMICS
     return __sync_add_and_fetch(&logger_gid, 1);
 #elif defined(__sun)
@@ -101,6 +102,18 @@ static uint64_t logger_get_gid(void) {
 #endif
 }
 
+void logger_set_gid(uint64_t gid) {
+#ifdef HAVE_GCC_64ATOMICS
+    __sync_add_and_fetch(&logger_gid, gid);
+#elif defined(__sun)
+    atomic_add_64(&logger_gid);
+#else
+    mutex_lock(&logger_atomics_mutex);
+    logger_gid = gid;
+    mutex_unlock(&logger_atomics_mutex);
+#endif
+}
+
 /* TODO: genericize lists. would be nice to import queue.h if the impact is
  * studied... otherwise can just write a local one.
  */
@@ -173,44 +186,44 @@ static void logger_set_flags(void) {
 static int _logger_thread_parse_ise(logentry *e, char *scratch) {
     int total;
     const char *cmd = "na";
-    char keybuf[KEY_MAX_LENGTH * 3 + 1];
+    char keybuf[KEY_MAX_URI_ENCODED_LENGTH];
     struct logentry_item_store *le = (struct logentry_item_store *) e->data;
     const char * const status_map[] = {
         "not_stored", "stored", "exists", "not_found", "too_large", "no_memory" };
     const char * const cmd_map[] = {
         "null", "add", "set", "replace", "append", "prepend", "cas" };
 
-    if (le->cmd <= 5)
+    if (le->cmd <= 6)
         cmd = cmd_map[le->cmd];
 
-    uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
+    uriencode(le->key, keybuf, le->nkey, KEY_MAX_URI_ENCODED_LENGTH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
-            "ts=%d.%d gid=%llu type=item_store key=%s status=%s cmd=%s ttl=%u clsid=%u\n",
+            "ts=%d.%d gid=%llu type=item_store key=%s status=%s cmd=%s ttl=%u clsid=%u cfd=%d\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
-            keybuf, status_map[le->status], cmd, le->ttl, le->clsid);
+            keybuf, status_map[le->status], cmd, le->ttl, le->clsid, le->sfd);
     return total;
 }
 
 static int _logger_thread_parse_ige(logentry *e, char *scratch) {
     int total;
     struct logentry_item_get *le = (struct logentry_item_get *) e->data;
-    char keybuf[KEY_MAX_LENGTH * 3 + 1];
+    char keybuf[KEY_MAX_URI_ENCODED_LENGTH];
     const char * const was_found_map[] = {
         "not_found", "found", "flushed", "expired" };
 
-    uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
+    uriencode(le->key, keybuf, le->nkey, KEY_MAX_URI_ENCODED_LENGTH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
-            "ts=%d.%d gid=%llu type=item_get key=%s status=%s clsid=%u\n",
+            "ts=%d.%d gid=%llu type=item_get key=%s status=%s clsid=%u cfd=%d\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
-            keybuf, was_found_map[le->was_found], le->clsid);
+            keybuf, was_found_map[le->was_found], le->clsid, le->sfd);
     return total;
 }
 
 static int _logger_thread_parse_ee(logentry *e, char *scratch) {
     int total;
-    char keybuf[KEY_MAX_LENGTH * 3 + 1];
+    char keybuf[KEY_MAX_URI_ENCODED_LENGTH];
     struct logentry_eviction *le = (struct logentry_eviction *) e->data;
-    uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
+    uriencode(le->key, keybuf, le->nkey, KEY_MAX_URI_ENCODED_LENGTH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
             "ts=%d.%d gid=%llu type=eviction key=%s fetch=%s ttl=%lld la=%d clsid=%u\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
@@ -222,9 +235,9 @@ static int _logger_thread_parse_ee(logentry *e, char *scratch) {
 #ifdef EXTSTORE
 static int _logger_thread_parse_extw(logentry *e, char *scratch) {
     int total;
-    char keybuf[KEY_MAX_LENGTH * 3 + 1];
+    char keybuf[KEY_MAX_URI_ENCODED_LENGTH];
     struct logentry_ext_write *le = (struct logentry_ext_write *) e->data;
-    uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
+    uriencode(le->key, keybuf, le->nkey, KEY_MAX_URI_ENCODED_LENGTH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
             "ts=%d.%d gid=%llu type=extwrite key=%s fetch=%s ttl=%lld la=%d clsid=%u bucket=%u\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
@@ -443,7 +456,7 @@ static int logger_thread_poll_watchers(int force_poll, int watcher) {
          */
         if (watchers_pollfds[nfd].revents & POLLIN) {
             char buf[1];
-            int res = read(w->sfd, buf, 1);
+            int res = ((conn*)w->c)->read(w->c, buf, 1);
             if (res == 0 || (res == -1 && (errno != EAGAIN && errno != EWOULDBLOCK))) {
                 L_DEBUG("LOGGER: watcher closed remotely\n");
                 logger_thread_close_watcher(w);
@@ -464,7 +477,7 @@ static int logger_thread_poll_watchers(int force_poll, int watcher) {
                         total = fwrite(data, 1, data_size, stderr);
                         break;
                     case LOGGER_WATCHER_CLIENT:
-                        total = write(w->sfd, data, data_size);
+                        total = ((conn*)w->c)->write(w->c, data, data_size);
                         break;
                 }
 
@@ -504,6 +517,8 @@ static void logger_thread_sum_stats(struct logger_stats *ls) {
 static void *logger_thread(void *arg) {
     useconds_t to_sleep = MIN_LOGGER_SLEEP;
     L_DEBUG("LOGGER: Starting logger thread\n");
+    // TODO: If we ever have item references in the logger code, will need to
+    // ensure everything is dequeued before stopping the thread.
     while (do_run_logger_thread) {
         int found_logs = 0;
         logger *l;
@@ -552,12 +567,11 @@ static int start_logger_thread(void) {
     return 0;
 }
 
-// future.
-/*static int stop_logger_thread(void) {
+static int stop_logger_thread(void) {
     do_run_logger_thread = 0;
     pthread_join(logger_tid, NULL);
     return 0;
-}*/
+}
 
 /*************************
  * Public functions for submitting logs and starting loggers from workers.
@@ -577,19 +591,16 @@ void logger_init(void) {
         abort();
     }
 
-    /* This can be removed once the global stats initializer is improved */
-    STATS_LOCK();
-    stats.log_worker_dropped = 0;
-    stats.log_worker_written = 0;
-    stats.log_watcher_skipped = 0;
-    stats.log_watcher_sent = 0;
-    STATS_UNLOCK();
     /* This is what adding a STDERR watcher looks like. should replace old
      * "verbose" settings. */
     //logger_add_watcher(NULL, 0);
     return;
 }
 
+void logger_stop(void) {
+    stop_logger_thread();
+}
+
 /* called *from* the thread using a logger.
  * initializes the per-thread bipbuf, links it into the list of loggers
  */
@@ -649,20 +660,20 @@ static void _logger_log_ext_write(logentry *e, item *it, uint8_t bucket) {
  * might be useful to store/print the flags an item has?
  * could also collapse this and above code into an "item status" struct. wait
  * for more endpoints to be written before making it generic, though.
- * TODO: This and below should track and reprint the client fd.
  */
 static void _logger_log_item_get(logentry *e, const int was_found, const char *key,
-        const int nkey, const uint8_t clsid) {
+        const int nkey, const uint8_t clsid, const int sfd) {
     struct logentry_item_get *le = (struct logentry_item_get *) e->data;
     le->was_found = was_found;
     le->nkey = nkey;
     le->clsid = clsid;
     memcpy(le->key, key, nkey);
+    le->sfd = sfd;
     e->size = sizeof(struct logentry_item_get) + nkey;
 }
 
 static void _logger_log_item_store(logentry *e, const enum store_item_type status,
-        const int comm, char *key, const int nkey, rel_time_t ttl, const uint8_t clsid) {
+        const int comm, char *key, const int nkey, rel_time_t ttl, const uint8_t clsid, int sfd) {
     struct logentry_item_store *le = (struct logentry_item_store *) e->data;
     le->status = status;
     le->cmd = comm;
@@ -674,6 +685,7 @@ static void _logger_log_item_store(logentry *e, const enum store_item_type statu
         le->ttl = 0;
     }
     memcpy(le->key, key, nkey);
+    le->sfd = sfd;
     e->size = sizeof(struct logentry_item_store) + nkey;
 }
 
@@ -740,7 +752,8 @@ enum logger_ret_type logger_log(logger *l, const enum log_entry_type event, cons
             char *key = va_arg(ap, char *);
             size_t nkey = va_arg(ap, size_t);
             uint8_t gclsid = va_arg(ap, int);
-            _logger_log_item_get(e, was_found, key, nkey, gclsid);
+            int gsfd = va_arg(ap, int);
+            _logger_log_item_get(e, was_found, key, nkey, gclsid, gsfd);
             va_end(ap);
             break;
         case LOGGER_ITEM_STORE_ENTRY:
@@ -751,7 +764,8 @@ enum logger_ret_type logger_log(logger *l, const enum log_entry_type event, cons
             size_t snkey = va_arg(ap, size_t);
             rel_time_t sttl = va_arg(ap, rel_time_t);
             uint8_t sclsid = va_arg(ap, int);
-            _logger_log_item_store(e, status, comm, skey, snkey, sttl, sclsid);
+            int ssfd = va_arg(ap, int);
+            _logger_log_item_store(e, status, comm, skey, snkey, sttl, sclsid, ssfd);
             va_end(ap);
             break;
     }
@@ -791,6 +805,7 @@ enum logger_add_watcher_ret logger_add_watcher(void *c, const int sfd, uint16_t
     logger_watcher *w = NULL;
     pthread_mutex_lock(&logger_stack_lock);
     if (watcher_count >= WATCHER_LIMIT) {
+        pthread_mutex_unlock(&logger_stack_lock);
         return LOGGER_ADD_WATCHER_TOO_MANY;
     }
 
