@@ -84,6 +84,7 @@ struct logentry_item_get {
     uint8_t was_found;
     uint8_t nkey;
     uint8_t clsid;
+    int sfd;
     char key[];
 };
 
@@ -93,19 +94,23 @@ struct logentry_item_store {
     rel_time_t ttl;
     uint8_t nkey;
     uint8_t clsid;
+    int sfd;
     char key[];
 };
 
 /* end intermediary structures */
 
+/* WARNING: cuddled items aren't compatible with warm restart. more code
+ * necessary to ensure log streams are all flushed/processed before stopping
+ */
 typedef struct _logentry {
     enum log_entry_subtype event;
+    uint8_t pad;
     uint16_t eflags;
     uint64_t gid;
     struct timeval tv; /* not monotonic! */
     int size;
     union {
-        void *entry; /* probably an item */
         char end;
     } data[];
 } logentry;
@@ -162,6 +167,7 @@ extern pthread_key_t logger_key;
 /* public functions */
 
 void logger_init(void);
+void logger_stop(void);
 logger *logger_create(void);
 
 #define LOGGER_LOG(l, flag, type, ...) \
@@ -183,4 +189,8 @@ enum logger_add_watcher_ret {
 
 enum logger_add_watcher_ret logger_add_watcher(void *c, const int sfd, uint16_t f);
 
+/* functions used by restart system */
+uint64_t logger_get_gid(void);
+void logger_set_gid(uint64_t gid);
+
 #endif
