@@ -20,17 +20,81 @@
 
 #include "config.h"
 #include "cache.h"
+#include "hash.h"
+#include "stats_prefix.h"
 #include "util.h"
 #include "protocol_binary.h"
+#ifdef TLS
+#include <openssl/ssl.h>
+#endif
 
 #define TMP_TEMPLATE "/tmp/test_file.XXXXXXX"
 
 enum test_return { TEST_SKIP, TEST_PASS, TEST_FAIL };
 
+struct conn {
+    int sock;
+#ifdef TLS
+    SSL_CTX   *ssl_ctx;
+    SSL    *ssl;
+#endif
+    ssize_t (*read)(struct conn  *c, void *buf, size_t count);
+    ssize_t (*write)(struct conn *c, const void *buf, size_t count);
+};
+
+struct settings {
+    char *hash_algorithm;
+};
+struct settings settings;
+
+static ssize_t tcp_read(struct conn *c, void *buf, size_t count);
+static ssize_t tcp_write(struct conn *c, const void *buf, size_t count);
+#ifdef TLS
+static ssize_t ssl_read(struct conn *c, void *buf, size_t count);
+static ssize_t ssl_write(struct conn *c, const void *buf, size_t count);
+#endif
+
+ssize_t tcp_read(struct conn *c, void *buf, size_t count) {
+    assert(c != NULL);
+    return read(c->sock, buf, count);
+}
+
+ssize_t tcp_write(struct conn *c, const void *buf, size_t count) {
+    assert(c != NULL);
+    return write(c->sock, buf, count);
+}
+#ifdef TLS
+ssize_t ssl_read(struct conn *c, void *buf, size_t count) {
+    assert(c != NULL);
+    return SSL_read(c->ssl, buf, count);
+}
+
+ssize_t ssl_write(struct conn *c, const void *buf, size_t count) {
+    assert(c != NULL);
+    return SSL_write(c->ssl, buf, count);
+}
+#endif
+
 static pid_t server_pid;
 static in_port_t port;
-static int sock;
+static struct conn *con = NULL;
 static bool allow_closed_read = false;
+static bool enable_ssl = false;
+
+static void close_conn() {
+    if (con == NULL) return;
+#ifdef TLS
+    if (con->ssl) {
+        SSL_shutdown(con->ssl);
+        SSL_free(con->ssl);
+    }
+    if (con->ssl_ctx)
+        SSL_CTX_free(con->ssl_ctx);
+#endif
+    if (con->sock > 0) close(con->sock);
+    free(con);
+    con = NULL;
+}
 
 static enum test_return cache_create_test(void)
 {
@@ -182,6 +246,138 @@ static enum test_return cache_redzone_test(void)
 #endif
 }
 
+static enum test_return test_stats_prefix_find(void) {
+    PREFIX_STATS *pfs1, *pfs2;
+
+    stats_prefix_clear();
+    pfs1 = stats_prefix_find("abc", 3);
+    assert(pfs1 == NULL);
+    pfs1 = stats_prefix_find("abc|", 4);
+    assert(pfs1 == NULL);
+
+    pfs1 = stats_prefix_find("abc:", 4);
+    assert(pfs1 != NULL);
+    assert(0ULL == (pfs1->num_gets + pfs1->num_sets + pfs1->num_deletes + pfs1->num_hits));
+    pfs2 = stats_prefix_find("abc:", 4);
+    assert(pfs1 == pfs2);
+    pfs2 = stats_prefix_find("abc:d", 5);
+    assert(pfs1 == pfs2);
+    pfs2 = stats_prefix_find("xyz123:", 6);
+    assert(pfs1 != pfs2);
+    pfs2 = stats_prefix_find("ab:", 3);
+    assert(pfs1 != pfs2);
+    return TEST_PASS;
+}
+
+static enum test_return test_stats_prefix_record_get(void) {
+    PREFIX_STATS *pfs;
+    stats_prefix_clear();
+
+    stats_prefix_record_get("abc:123", 7, false);
+    pfs = stats_prefix_find("abc:123", 7);
+    assert(1 == pfs->num_gets);
+    assert(0 == pfs->num_hits);
+    stats_prefix_record_get("abc:456", 7, false);
+    assert(2 == pfs->num_gets);
+    assert(0 == pfs->num_hits);
+    stats_prefix_record_get("abc:456", 7, true);
+    assert(3 == pfs->num_gets);
+    assert(1 == pfs->num_hits);
+    stats_prefix_record_get("def:", 4, true);
+    assert(3 == pfs->num_gets);
+    assert(1 == pfs->num_hits);
+    return TEST_PASS;
+}
+
+static enum test_return test_stats_prefix_record_delete(void) {
+    PREFIX_STATS *pfs;
+    stats_prefix_clear();
+
+    stats_prefix_record_delete("abc:123", 7);
+    pfs = stats_prefix_find("abc:123", 7);
+    assert(0 == pfs->num_gets);
+    assert(0 == pfs->num_hits);
+    assert(1 == pfs->num_deletes);
+    assert(0 == pfs->num_sets);
+    stats_prefix_record_delete("def:", 4);
+    assert(1 == pfs->num_deletes);
+    return TEST_PASS;
+}
+
+static enum test_return test_stats_prefix_record_set(void) {
+    PREFIX_STATS *pfs;
+    stats_prefix_clear();
+
+    stats_prefix_record_set("abc:123", 7);
+    pfs = stats_prefix_find("abc:123", 7);
+    assert(0 == pfs->num_gets);
+    assert(0 == pfs->num_hits);
+    assert(0 == pfs->num_deletes);
+    assert(1 == pfs->num_sets);
+    stats_prefix_record_delete("def:", 4);
+    assert(1 == pfs->num_sets);
+    return TEST_PASS;
+}
+
+static enum test_return test_stats_prefix_dump(void) {
+    int hashval = hash("abc", 3) % PREFIX_HASH_SIZE;
+    char tmp[500];
+    char *expected;
+    int keynum;
+    int length;
+
+    stats_prefix_clear();
+
+    assert(strcmp("END\r\n", stats_prefix_dump(&length)) == 0);
+    assert(5 == length);
+    stats_prefix_record_set("abc:123", 7);
+    expected = "PREFIX abc get 0 hit 0 set 1 del 0\r\nEND\r\n";
+    assert(strcmp(expected, stats_prefix_dump(&length)) == 0);
+    assert(strlen(expected) == length);
+    stats_prefix_record_get("abc:123", 7, false);
+    expected = "PREFIX abc get 1 hit 0 set 1 del 0\r\nEND\r\n";
+    assert(strcmp(expected, stats_prefix_dump(&length)) == 0);
+    assert(strlen(expected) == length);
+    stats_prefix_record_get("abc:123", 7, true);
+    expected = "PREFIX abc get 2 hit 1 set 1 del 0\r\nEND\r\n";
+    assert(strcmp(expected, stats_prefix_dump(&length)) == 0);
+    assert(strlen(expected) == length);
+    stats_prefix_record_delete("abc:123", 7);
+    expected = "PREFIX abc get 2 hit 1 set 1 del 1\r\nEND\r\n";
+    assert(strcmp(expected, stats_prefix_dump(&length)) == 0);
+    assert(strlen(expected) == length);
+
+    /* The order of results might change if we switch hash functions. */
+    stats_prefix_record_delete("def:123", 7);
+    expected = "PREFIX abc get 2 hit 1 set 1 del 1\r\n"
+               "PREFIX def get 0 hit 0 set 0 del 1\r\n"
+               "END\r\n";
+    assert(strcmp(expected, stats_prefix_dump(&length)) == 0);
+    assert(strlen(expected) == length);
+
+    /* Find a key that hashes to the same bucket as "abc" */
+    bool found_match = false;
+    for (keynum = 0; keynum < PREFIX_HASH_SIZE * 100; keynum++) {
+        snprintf(tmp, sizeof(tmp), "%d:", keynum);
+        /* -1 because only the prefix portion is used when hashing */
+        if (hashval == hash(tmp, strlen(tmp) - 1) % PREFIX_HASH_SIZE) {
+            found_match = true;
+            break;
+        }
+    }
+    assert(found_match);
+    stats_prefix_record_set(tmp, strlen(tmp));
+    snprintf(tmp, sizeof(tmp),
+             "PREFIX %d get 0 hit 0 set 1 del 0\r\n"
+             "PREFIX abc get 2 hit 1 set 1 del 1\r\n"
+             "PREFIX def get 0 hit 0 set 0 del 1\r\n"
+             "END\r\n", keynum);
+    assert(strcmp(tmp, stats_prefix_dump(&length)) == 0);
+    assert(strlen(tmp) == length);
+
+    return TEST_PASS;
+}
+
 static enum test_return test_safe_strtoul(void) {
     uint32_t val;
     assert(safe_strtoul("123", &val));
@@ -318,7 +514,7 @@ static pid_t start_server(in_port_t *port_out, bool daemon, int timeout) {
 
     if (pid == 0) {
         /* Child */
-        char *argv[20];
+        char *argv[24];
         int arg = 0;
         char tmo[24];
         snprintf(tmo, sizeof(tmo), "%u", timeout);
@@ -339,6 +535,15 @@ static pid_t start_server(in_port_t *port_out, bool daemon, int timeout) {
         argv[arg++] = "-1";
         argv[arg++] = "-U";
         argv[arg++] = "0";
+#ifdef TLS
+        if (enable_ssl) {
+            argv[arg++] = "-Z";
+            argv[arg++] = "-o";
+            argv[arg++] = "ssl_chain_cert=t/server_crt.pem";
+            argv[arg++] = "-o";
+            argv[arg++] = "ssl_key=t/server_key.pem";
+        }
+#endif
         /* Handle rpmbuild and the like doing this as root */
         if (getuid() == 0) {
             argv[arg++] = "-u";
@@ -445,8 +650,16 @@ static struct addrinfo *lookuphost(const char *hostname, in_port_t port)
     return ai;
 }
 
-static int connect_server(const char *hostname, in_port_t port, bool nonblock)
+static struct conn *connect_server(const char *hostname, in_port_t port,
+                            bool nonblock, const bool ssl)
 {
+    struct conn *c;
+    if (!(c = (struct conn *)calloc(1, sizeof(struct conn)))) {
+        fprintf(stderr, "Failed to allocate the client connection: %s\n",
+                strerror(errno));
+        return NULL;
+    }
+
     struct addrinfo *ai = lookuphost(hostname, port);
     int sock = -1;
     if (ai != NULL) {
@@ -472,12 +685,49 @@ static int connect_server(const char *hostname, in_port_t port, bool nonblock)
 
        freeaddrinfo(ai);
     }
-    return sock;
+    c->sock = sock;
+#ifdef TLS
+    if (sock > 0 && ssl) {
+        c->ssl_ctx = SSL_CTX_new(SSLv23_client_method());
+        if (c->ssl_ctx == NULL) {
+            fprintf(stderr, "Failed to create the SSL context: %s\n",
+                strerror(errno));
+            close(sock);
+            sock = -1;
+        }
+        c->ssl = SSL_new(c->ssl_ctx);
+        if (c->ssl == NULL) {
+            fprintf(stderr, "Failed to create the SSL object: %s\n",
+                strerror(errno));
+            close(sock);
+            sock = -1;
+        }
+        SSL_set_fd (c->ssl, c->sock);
+        int ret = SSL_connect(c->ssl);
+        if (ret < 0) {
+            int err = SSL_get_error(c->ssl, ret);
+            if (err == SSL_ERROR_SYSCALL || err == SSL_ERROR_SSL) {
+                fprintf(stderr, "SSL connection failed with error code : %s\n",
+                    strerror(errno));
+                close(sock);
+                sock = -1;
+            }
+        }
+        c->read = ssl_read;
+        c->write = ssl_write;
+    } else
+#endif
+    {
+        c->read = tcp_read;
+        c->write = tcp_write;
+    }
+    return c;
 }
 
 static enum test_return test_vperror(void) {
     int rv = 0;
     int oldstderr = dup(STDERR_FILENO);
+    assert(oldstderr >= 0);
     char tmpl[sizeof(TMP_TEMPLATE)+1];
     strncpy(tmpl, TMP_TEMPLATE, sizeof(TMP_TEMPLATE)+1);
 
@@ -525,7 +775,7 @@ static void send_ascii_command(const char *buf) {
     size_t len = strlen(buf);
 
     do {
-        ssize_t nw = write(sock, ptr + offset, len - offset);
+        ssize_t nw = con->write((void*)con, ptr + offset, len - offset);
         if (nw == -1) {
             if (errno != EINTR) {
                 fprintf(stderr, "Failed to write: %s\n", strerror(errno));
@@ -547,7 +797,7 @@ static void read_ascii_response(char *buffer, size_t size) {
     off_t offset = 0;
     bool need_more = true;
     do {
-        ssize_t nr = read(sock, buffer + offset, 1);
+        ssize_t nr = con->read(con, buffer + offset, 1);
         if (nr == -1) {
             if (errno != EINTR) {
                 fprintf(stderr, "Failed to read: %s\n", strerror(errno));
@@ -568,10 +818,12 @@ static void read_ascii_response(char *buffer, size_t size) {
 static enum test_return test_issue_92(void) {
     char buffer[1024];
 
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     send_ascii_command("stats cachedump 1 0 0\r\n");
+
     read_ascii_response(buffer, sizeof(buffer));
     assert(strncmp(buffer, "END", strlen("END")) == 0);
 
@@ -579,8 +831,9 @@ static enum test_return test_issue_92(void) {
     read_ascii_response(buffer, sizeof(buffer));
     assert(strncmp(buffer, "CLIENT_ERROR", strlen("CLIENT_ERROR")) == 0);
 
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
     return TEST_PASS;
 }
 
@@ -589,14 +842,17 @@ static enum test_return test_issue_102(void) {
     memset(buffer, ' ', sizeof(buffer));
     buffer[sizeof(buffer) - 1] = '\0';
 
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     send_ascii_command(buffer);
     /* verify that the server closed the connection */
-    assert(read(sock, buffer, sizeof(buffer)) == 0);
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    assert(con->read(con, buffer, sizeof(buffer)) == 0);
+
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     snprintf(buffer, sizeof(buffer), "gets ");
     size_t offset = 5;
@@ -625,22 +881,25 @@ static enum test_return test_issue_102(void) {
     buffer[sizeof(buffer) - 1] = '\0';
     send_ascii_command(buffer);
     /* verify that the server closed the connection */
-    assert(read(sock, buffer, sizeof(buffer)) == 0);
+    assert(con->read(con, buffer, sizeof(buffer)) == 0);
 
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     return TEST_PASS;
 }
 
 static enum test_return start_memcached_server(void) {
     server_pid = start_server(&port, false, 600);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
     return TEST_PASS;
 }
 
 static enum test_return stop_memcached_server(void) {
-    close(sock);
+    close_conn();
     if (server_pid != -1) {
         assert(kill(server_pid, SIGTERM) == 0);
     }
@@ -651,14 +910,15 @@ static enum test_return stop_memcached_server(void) {
 static enum test_return shutdown_memcached_server(void) {
     char buffer[1024];
 
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     send_ascii_command("shutdown\r\n");
     /* verify that the server closed the connection */
-    assert(read(sock, buffer, sizeof(buffer)) == 0);
+    assert(con->read(con, buffer, sizeof(buffer)) == 0);
 
-    close(sock);
+    close_conn();
 
     /* We set server_pid to -1 so that we don't later call kill() */
     if (kill(server_pid, 0) == 0) {
@@ -694,8 +954,7 @@ static void safe_send(const void* buf, size_t len, bool hickup)
                 num_bytes = (rand() % 1023) + 1;
             }
         }
-
-        ssize_t nw = write(sock, ptr + offset, num_bytes);
+        ssize_t nw = con->write(con, ptr + offset, num_bytes);
         if (nw == -1) {
             if (errno != EINTR) {
                 fprintf(stderr, "Failed to write: %s\n", strerror(errno));
@@ -716,7 +975,7 @@ static bool safe_recv(void *buf, size_t len) {
     }
     off_t offset = 0;
     do {
-        ssize_t nr = read(sock, ((char*)buf) + offset, len - offset);
+        ssize_t nr = con->read(con, ((char*)buf) + offset, len - offset);
         if (nr == -1) {
             if (errno != EINTR) {
                 fprintf(stderr, "Failed to read: %s\n", strerror(errno));
@@ -1064,9 +1323,10 @@ static enum test_return test_binary_quit_impl(uint8_t cmd) {
     }
 
     /* Socket should be closed now, read should return 0 */
-    assert(read(sock, buffer.bytes, sizeof(buffer.bytes)) == 0);
-    close(sock);
-    sock = connect_server("127.0.0.1", port, false);
+    assert(con->read(con, buffer.bytes, sizeof(buffer.bytes)) == 0);
+    close_conn();
+    con = connect_server("127.0.0.1", port, false, enable_ssl);
+    assert(con);
 
     return TEST_PASS;
 }
@@ -1842,6 +2102,7 @@ static enum test_return test_binary_pipeline_hickup(void)
     if ((ret = pthread_create(&tid, NULL,
                               binary_hickup_recv_verification_thread, NULL)) != 0) {
         fprintf(stderr, "Can't create thread: %s\n", strerror(ret));
+        free(buffer);
         return TEST_FAIL;
     }
 
@@ -1867,7 +2128,7 @@ static enum test_return test_binary_pipeline_hickup(void)
 static enum test_return test_issue_101(void) {
     enum { max = 2 };
     enum test_return ret = TEST_PASS;
-    int fds[max];
+    struct conn *conns[max];
     int ii = 0;
     pid_t child = 0;
 
@@ -1881,15 +2142,17 @@ static enum test_return test_issue_101(void) {
     server_pid = start_server(&port, false, 1000);
 
     for (ii = 0; ii < max; ++ii) {
-        fds[ii] = connect_server("127.0.0.1", port, true);
-        assert(fds[ii] > 0);
+        conns[ii] = NULL;
+        conns[ii] = connect_server("127.0.0.1", port, true, enable_ssl);
+        assert(conns[ii]);
+        assert(conns[ii]->sock > 0);
     }
 
     /* Send command on the connection until it blocks */
     for (ii = 0; ii < max; ++ii) {
         bool more = true;
         do {
-            ssize_t err = write(fds[ii], command, cmdlen);
+            ssize_t err = conns[ii]->write(conns[ii], command, cmdlen);
             if (err == -1) {
                 switch (errno) {
                 case EINTR:
@@ -1916,16 +2179,28 @@ static enum test_return test_issue_101(void) {
         assert(c == child);
         assert(stat == 0);
     } else {
-        sock = connect_server("127.0.0.1", port, false);
+        con = connect_server("127.0.0.1", port, false, enable_ssl);
+        assert(con);
         ret = test_binary_noop();
-        close(sock);
         exit(0);
     }
 
  cleanup:
     /* close all connections */
     for (ii = 0; ii < max; ++ii) {
-        close(fds[ii]);
+        struct conn *c = conns[ii];
+        if (c == NULL) continue;
+#ifdef TLS
+        if (c->ssl) {
+            SSL_shutdown(c->ssl);
+            SSL_free(c->ssl);
+        }
+        if (c->ssl_ctx)
+            SSL_CTX_free(c->ssl_ctx);
+#endif
+        if (c->sock > 0) close(c->sock);
+        free(conns[ii]);
+        conns[ii] = NULL;
     }
 
     assert(kill(server_pid, SIGTERM) == 0);
@@ -1946,6 +2221,11 @@ struct testcase testcases[] = {
     { "cache_destructor", cache_destructor_test },
     { "cache_reuse", cache_reuse_test },
     { "cache_redzone", cache_redzone_test },
+    { "stats_prefix_find", test_stats_prefix_find },
+    { "stats_prefix_record_get", test_stats_prefix_record_get },
+    { "stats_prefix_record_delete", test_stats_prefix_record_delete },
+    { "stats_prefix_record_set", test_stats_prefix_record_set },
+    { "stats_prefix_dump", test_stats_prefix_dump },
     { "issue_161", test_issue_161 },
     { "strtol", test_safe_strtol },
     { "strtoll", test_safe_strtoll },
@@ -1996,10 +2276,28 @@ struct testcase testcases[] = {
     { NULL, NULL }
 };
 
+/* Stub out function defined in memcached.c */
+void STATS_LOCK(void);
+void STATS_UNLOCK(void);
+void STATS_LOCK(void)
+{}
+void STATS_UNLOCK(void)
+{}
+
 int main(int argc, char **argv)
 {
     int exitcode = 0;
     int ii = 0, num_cases = 0;
+#ifdef TLS
+    if (getenv("SSL_TEST") != NULL) {
+        SSLeay_add_ssl_algorithms();
+        SSL_load_error_strings();
+        enable_ssl = true;
+    }
+#endif
+    /* Stats prefix test is sensitive to the choice of hash function */
+    hash_init(JENKINS_HASH);
+    stats_prefix_init(':');
 
     for (num_cases = 0; testcases[num_cases].description; num_cases++) {
         /* Just counting */
