@@ -17,6 +17,10 @@
 #include <atomic.h>
 #endif
 
+#ifdef TLS
+#include <openssl/ssl.h>
+#endif
+
 #define ITEMS_PER_ALLOC 64
 
 /* An item in the connection queue. */
@@ -33,6 +37,7 @@ struct conn_queue_item {
     enum network_transport     transport;
     enum conn_queue_item_modes mode;
     conn *c;
+    void    *ssl;
     CQ_ITEM          *next;
 };
 
@@ -184,6 +189,60 @@ void pause_threads(enum pause_thread_types type) {
     pthread_mutex_unlock(&init_lock);
 }
 
+// MUST not be called with any deeper locks held
+// MUST be called only by parent thread
+// Note: listener thread is the "main" event base, which has exited its
+// loop in order to call this function.
+void stop_threads(void) {
+    char buf[1];
+    int i;
+
+    // assoc can call pause_threads(), so we have to stop it first.
+    stop_assoc_maintenance_thread();
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped assoc\n");
+
+    if (settings.verbose > 0)
+        fprintf(stderr, "asking workers to stop\n");
+    buf[0] = 's';
+    pthread_mutex_lock(&init_lock);
+    init_count = 0;
+    for (i = 0; i < settings.num_threads; i++) {
+        if (write(threads[i].notify_send_fd, buf, 1) != 1) {
+            perror("Failed writing to notify pipe");
+            /* TODO: This is a fatal problem. Can it ever happen temporarily? */
+        }
+    }
+    wait_for_thread_registration(settings.num_threads);
+    pthread_mutex_unlock(&init_lock);
+
+    if (settings.verbose > 0)
+        fprintf(stderr, "asking background threads to stop\n");
+
+    // stop each side thread.
+    // TODO: Verify these all work if the threads are already stopped
+    stop_item_crawler_thread(CRAWLER_WAIT);
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped lru crawler\n");
+    stop_lru_maintainer_thread();
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped maintainer\n");
+    stop_slab_maintenance_thread();
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped slab mover\n");
+    logger_stop();
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped logger thread\n");
+    stop_conn_timeout_thread();
+    if (settings.verbose > 0)
+        fprintf(stderr, "stopped idle timeout thread\n");
+
+    if (settings.verbose > 0)
+        fprintf(stderr, "all background threads stopped\n");
+
+    // At this point, every background thread must be stopped.
+}
+
 /*
  * Initializes a connection queue.
  */
@@ -361,6 +420,15 @@ static void setup_thread(LIBEVENT_THREAD *me) {
         exit(EXIT_FAILURE);
     }
 #endif
+#ifdef TLS
+    if (settings.ssl_enabled) {
+        me->ssl_wbuf = (char *)malloc((size_t)settings.ssl_wbuf_size);
+        if (me->ssl_wbuf == NULL) {
+            fprintf(stderr, "Failed to allocate the SSL write buffer\n");
+            exit(EXIT_FAILURE);
+        }
+    }
+#endif
 }
 
 /*
@@ -386,6 +454,9 @@ static void *worker_libevent(void *arg) {
 
     event_base_loop(me->base, 0);
 
+    // same mechanism used to watch for all threads exiting.
+    register_thread_initialized();
+
     event_base_free(me->base);
     return NULL;
 }
@@ -419,7 +490,7 @@ static void thread_libevent_process(int fd, short which, void *arg) {
             case queue_new_conn:
                 c = conn_new(item->sfd, item->init_state, item->event_flags,
                                    item->read_buffer_size, item->transport,
-                                   me->base);
+                                   me->base, item->ssl);
                 if (c == NULL) {
                     if (IS_UDP(item->transport)) {
                         fprintf(stderr, "Can't listen for events on UDP socket\n");
@@ -429,10 +500,22 @@ static void thread_libevent_process(int fd, short which, void *arg) {
                             fprintf(stderr, "Can't listen for events on fd %d\n",
                                 item->sfd);
                         }
+#ifdef TLS
+                        if (item->ssl) {
+                            SSL_shutdown(item->ssl);
+                            SSL_free(item->ssl);
+                        }
+#endif
                         close(item->sfd);
                     }
                 } else {
                     c->thread = me;
+#ifdef TLS
+                    if (settings.ssl_enabled && c->ssl != NULL) {
+                        assert(c->thread && c->thread->ssl_wbuf);
+                        c->ssl_wbuf = c->thread->ssl_wbuf;
+                    }
+#endif
                 }
                 break;
 
@@ -455,6 +538,10 @@ static void thread_libevent_process(int fd, short which, void *arg) {
         }
         conn_close_idle(conns[timeout_fd]);
         break;
+    /* asked to stop */
+    case 's':
+        event_base_loopexit(me->base, NULL);
+        break;
     }
 }
 
@@ -467,14 +554,14 @@ static int last_thread = -1;
  * of an incoming connection.
  */
 void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
-                       int read_buffer_size, enum network_transport transport) {
+                       int read_buffer_size, enum network_transport transport, void *ssl) {
     CQ_ITEM *item = cqi_new();
     char buf[1];
     if (item == NULL) {
         close(sfd);
         /* given that malloc failed this may also fail, but let's try */
         fprintf(stderr, "Failed to allocate memory for connection object\n");
-        return ;
+        return;
     }
 
     int tid = (last_thread + 1) % settings.num_threads;
@@ -489,10 +576,11 @@ void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
     item->read_buffer_size = read_buffer_size;
     item->transport = transport;
     item->mode = queue_new_conn;
+    item->ssl = ssl;
 
     cq_push(thread->new_conn_queue, item);
 
-    MEMCACHED_CONN_DISPATCH(sfd, thread->thread_id);
+    MEMCACHED_CONN_DISPATCH(sfd, (int64_t)thread->thread_id);
     buf[0] = 'c';
     if (write(thread->notify_send_fd, buf, 1) != 1) {
         perror("Writing to thread notify pipe");
@@ -531,6 +619,13 @@ void sidethread_conn_close(conn *c) {
     c->state = conn_closed;
     if (settings.verbose > 1)
         fprintf(stderr, "<%d connection closed from side thread.\n", c->sfd);
+#ifdef TLS
+    if (c->ssl) {
+        c->ssl_wbuf = NULL;
+        SSL_shutdown(c->ssl);
+        SSL_free(c->ssl);
+    }
+#endif
     close(c->sfd);
 
     STATS_LOCK();
@@ -566,6 +661,17 @@ item *item_get(const char *key, const size_t nkey, conn *c, const bool do_update
     return it;
 }
 
+// returns an item with the item lock held.
+// lock will still be held even if return is NULL, allowing caller to replace
+// an item atomically if desired.
+item *item_get_locked(const char *key, const size_t nkey, conn *c, const bool do_update, uint32_t *hv) {
+    item *it;
+    *hv = hash(key, nkey);
+    item_lock(*hv);
+    it = do_item_get(key, nkey, *hv, c, do_update);
+    return it;
+}
+
 item *item_touch(const char *key, size_t nkey, uint32_t exptime, conn *c) {
     item *it;
     uint32_t hv;
